import os

from Ft.Xml.Domlette import NonvalidatingReader
from Ft.Lib import Uri
import Ft.Xml.Domlette as dom
import warnings

# Parsing from a string will otherwise print a
# warning which I want to suppress.
def disable_warnings(*args): pass

warnings.filterwarnings("ignore", category=Warning)
warnings.showwarning = disable_warnings

# Used to store and retrieve information in a common format.
class XMLStorage(object):

    # The directory where information id stored
    ___DATADIR="data/"
    
    def __init__(self, datadir="data/"):
        """Create the XMLStorage. 
        
        Check if the dir is present, otherwise
        create it 
        """
        if not os.path.exists(datadir):
            os.mkdir(datadir)
        self.___DATADIR = datadir
            
    
    # @ReturnType void
    # @ParamType aColName string
    # Clear the given collection.
    def clearCollection(self, aColName):
        """Clear the given collection.
        
        Implementation is simple: just remove the file
        """
        if os.path.exists(self.___DATADIR + aColName):
            os.remove(self.___DATADIR + aColName)


    # @ReturnType void
    # @ParamType aColName string
    # @ParamType aXpath string
    # Remove a resource from the given collection.
    def removeResource(self, aColName, aXpath):
        """The XPath matches some elements, these are removed
        from the resource."""
        if not os.path.exists(self.___DATADIR + aColName):
            return
            
        file_uri = Uri.OsPathToUri(self.___DATADIR + aColName)
        doc = NonvalidatingReader.parseUri(file_uri)
        res = doc.xpath(aXpath)
        
        for e in res:
            p = e.parentNode
            p.removeChild(e)
        
        FILE = open(self.___DATADIR + aColName, "w")
        dom.Print(doc, FILE)
        FILE.close()

    # @ParamType aColName string
    # @ParamType aXpath string
    # Retrieve the matching xpath in the given collection.
    def retrieve(self, aColName, aXpath):
        """Retrieve the given XPath expression from the document"""
        if not os.path.exists(self.___DATADIR + aColName):
            return None
        file_uri = Uri.OsPathToUri(self.___DATADIR + aColName)
        doc = NonvalidatingReader.parseUri(file_uri)
        return doc.xpath(aXpath)

    # @ReturnType void
    # @ParamType aColName string
    # @ParamType aSnippet string
    # Store an xml snippet in the given collection.
    def store(self, aColName, aSnippet):
        """Store a given snippet in the collection."""
        if not os.path.exists(self.___DATADIR + aColName):
            FILE = open(self.___DATADIR + aColName, "w")
            doc = dom.implementation.createDocument(None, "collection", None)
            dom.Print(doc, FILE)
            FILE.close()
        else:
            # Create a doc from the file:
            file_uri = Uri.OsPathToUri(self.___DATADIR + aColName)
            doc = NonvalidatingReader.parseUri(file_uri)
        
        # Doc should be alright now
        root = doc.documentElement
        snippet = NonvalidatingReader.parseString(aSnippet)
        root.appendChild(snippet.documentElement)
        FILE = open(self.___DATADIR + aColName, "w")
        dom.Print(root, FILE)
        FILE.close()
        
        