import sys, os, marshal
import copy

from VariableGenerator import *
from rubacon.utility.helpers import get_attr_list_from_model, get_attrs_from_class
from networkx import *
from rubacon.parsers.RuleParser import RuleParser


# Generate Prolog for verifying the constraints defined in the storage.
class PrologGenerator(object):

    ___MODEL_STORAGE = "data/db.model"
    ___INFO_STORAGE = "data/db.information"

        
    # @ParamType aFilename 
    # Generate the code and write it to the given file.
    def generate(self, aFilename, config=None):
        """Generate the prolog code and write it to the given file."""
        
        if config == None:
            config = ConfigParser.ConfigParser()
            config.readfp(open('rubacon/defaults.cfg'))
            logging.debug("Created config parser in PrologGenerator.")
            
        self.___INFO_STORAGE = config.get('information', 'storage')
        self.___MODEL_STORAGE = config.get('model', 'storage')
        
        vg = unique_variable_generator

        # Preload the model and info and open the file
        # for writing.
        inf = open(self.___INFO_STORAGE)
        a = marshal.load(inf)
        inf.close()
        inf = open(self.___MODEL_STORAGE)
        m = marshal.load(inf)
        inf.close()
        FILE = open(aFilename, "w")
        
        # Write the atoms for each object first, e.g.:
        #     User('userName', 'name1')
        for className in a.iterkeys():
            for o in a[className]:
                first = True
                FILE.write(className.lower() + "(")
                keys = o.keys()
                keys.sort()
                for v in keys:
                    if not first:
                        FILE.write(", ")
                    FILE.write(o[v])
                    first = False
                FILE.write(").\n")
        
        # The connecting attributes must be generated
        # Get all dependencies:
        for d in m['dependencyDefs']:
            #print >> sys.stderr, d
            # get the client:
            for classDef in m['classDefs']:
                if classDef['name'] == d[0]:
                    client = classDef
                if classDef['name'] == d[1]:
                    supplier = classDef
            # print >> sys.stderr, "Connection: %s -> %s" % (supplier['name'], client['name'])
            ## Now write the connecting attribute:
            FILE.write("%s_%s(" % (supplier['name'].lower(), client['name'].lower()))
            supplier_pred = supplier['name'].lower() + "("
            client_pred = client['name'].lower() + "("
            # Get the supplier and write its attributes,
            # save the mapping into a hash
            mapping = {}
            first = True
            #print >> sys.stderr, "Supplier attrs"
            #print >> sys.stderr, supplier['attrs']
            #print >> sys.stderr, "Client attrs"
            #print >> sys.stderr, client['attrs']
            
            attrs = supplier['attrs']
            attrs.sort()
            for attr in attrs:
                v = vg.next()
                mapping[attr] = v
                if not first:
                    FILE.write(', ')
                    supplier_pred += ', '
                else:
                    first = False
                FILE.write(v)
                supplier_pred += v
            
            # FILE.write(' || ')
            # Now write the client attributes, but replace the ones with the same
            # name by the same variable
            first = True
            attrs = client['attrs']
            attrs.sort()
            for attr in attrs:
                # We aleady have a var fof this
                if attr in mapping:
                    v = mapping[attr]
                else:
                    v = vg.next()
                
                if not first:
                    client_pred += ', '
                else:
                    first = False
                FILE.write(', ')
                FILE.write(v)
                client_pred += v
            FILE.write(") :- %s), %s).\n" % (supplier_pred, client_pred))
        
        # Write out the predicates for the rules:
        rules = RuleParser.getRulesFromStore()
        # FIXME: add context mapping
        predicateResolver = {}
        alreadyInserted = []
        nameResolver = {}
        # print >> sys.stderr, rules 
        for rule in rules:
            # print >> sys.stderr, rule['subrules']
            for subrule in rule['subrules']:
                predicate = ""
                ruleBody = ""
                # print >> sys.stderr, rule
                s = Signature(m, head=subrule['head'], target=subrule['target'])
                bWrite = True
                if s.predicate in alreadyInserted:
                    bWrite = False   # This is different in the original code, chech if it works
                alreadyInserted.append(s.predicate)
                predicate += s.predicate.lower() + "("
                ruleBody += " :- "
                currentSig = Signature(m, predicate=s.path[0])
                previousPredicate = currentSig.predicate
                ruleBody += currentSig.predicate.lower() + "("
                lookupVarName = {}
                
                for i in range(0, currentSig.get_arg_count()):
                    vn = vg.next()
                    predicate += vn
                    ruleBody += vn
                    lookupVarName[currentSig.predicate + "." + currentSig.get_attribute_name(i)] = vn
                    predicate += ", "
                    
                    if (i < currentSig.get_arg_count() - 1):
                        ruleBody += ", "
                
                for node in s.path[1:]:
                    currentSig = Signature(m, predicate=node)
                    ruleBody += "), "
                    ruleBody += currentSig.predicate.lower() + "("
                    conName = self.__getConName(m, previousPredicate, currentSig.predicate)
                    # print >> sys.stderr, "ConName: %s" % conName
                    conAttrName = lookupVarName[previousPredicate + "." + conName]
                    # print >> sys.stderr, "ConAttrName: %s" % conAttrName
                    for i in range(0, currentSig.argCount):
                        currName = currentSig.get_attribute_name(i)
                        if conName == currName:
                            predicate += conAttrName
                            ruleBody += conAttrName
                            lookupVarName[currentSig.predicate + "." + currName] = conAttrName
                        else:
                            vn = vg.next()
                            predicate += vn
                            ruleBody += vn
                            lookupVarName[currentSig.predicate + "." + currName] = vn
                        if not (i == (currentSig.argCount - 1) and node == s.path[-1]):  
                            predicate += ", "
                        if i < currentSig.argCount - 1:
                            ruleBody += ", "
                    previousPredicate = currentSig.predicate
                
                predicate += ")"
                ruleBody += ")."
                predicateResolver[subrule['name']] = predicate
                
                # print >> sys.stderr, predicate + ruleBody
                if bWrite:
                    FILE.write(predicate + ruleBody + "\n")
                
                # The lookup for resolving names in constraints.
                lookupVarName['HEAD'] = subrule['head']
                lookupVarName['TARGET'] = subrule['target']
                # print >> sys.stderr, "nameRes: %s" % subrule['name']
                nameResolver[subrule['name']] = lookupVarName
            
            # Now write the rule itself:
            paramCount = len(rule['params'])
            ruleText = ""
            ruleText += rule['name']
            if paramCount > 0:
                ruleText += "("
            
            paramResolver = {}
            for i in range(0, paramCount):
                nName = vg.next()
                ruleText += nName
                paramResolver[rule['params'][i]] = nName
                # TODO: ContextMapping add mapping from PROLOG -> RuleName
                if (i < paramCount - 1):
                    ruleText += ", "
            if paramCount > 0:
                ruleText += ")"
            
            ruleText += " :- "
            for key in predicateResolver.keys():
                p = predicateResolver[key]
                ruleText += p
                if not key == predicateResolver.keys()[-1]:
                    ruleText += ", "
            added = 0
            for subrule in rule['subrules']:
                for constraint in subrule['constraints']:
                    for entry in constraint['entries']:
                        if added == 0:
                            ruleText += ", "
                        else:
                            ruleText += " = "
                            added = -1
                        if entry['type'] == 'parameter':
                            p = paramResolver[entry['name']]
                            ruleText += p
                            added += 1
                        elif entry['type'] == 'reference':
                            r = entry['name']
                            s = r.split(".")
                            # print >> sys.stderr, r
                            # print >> sys.stderr, nameResolver
                            res = nameResolver[s[0]]
                            if res == None:
                                r = subrule['name'] + "." + entry['name']
                                s = r.split(".")
                                res = nameResolver[s[0]]
                            if s[1] == 'HEAD' or s[1] == 'TARGET':
                                s[1] = res[s[1]]
                            search = s[1]
                            
                            for k in range(2, len(s)):
                                search += ("." + s[k])
                            
                            try:
                                n = res[search]
                                ruleText += res[search]
                            except:
                                print >> sys.stderr, "Element not found: %s" % search
                            
                            added += 1
                        elif entry['type'] == 'value':
                            ruleText += '"%s"' % entry['name']
                            added += 1
                        
                            
            # print >> sys.stderr, ruleText
            FILE.write(ruleText + ".\n")
        FILE.close()

    def __getConName(self, m, head, target):
        """Get the name of the connection from head to target"""
        for dep in m['dependencyDefs']:
            if dep[1] == head and dep[0] == target:
                return dep[2]
        return ""
#    def __findConnectingAttr(self, m, previous, current):
#        """Find the Attribute that connects the two predicates"""
#        neighbors = self.__getDepsForClass(m, previous)
#        for neighbor in neighbors:
            
            
#    def __getDepsForClass(self, m, name):
#        """Get all dependencies for a class"""
#        retVal = []
#        for dep in m['dependencyDefs']:
#            if dep[0] == name:
#                retval.append(dep[1])
#        return retVal
                    
class Signature(object):
    """Create a signature for the given head -> target values."""        
        
    def __init__(self, model, head=None, target=None, predicate=None):
        """Create a new Signature from head to target"""
        if predicate == None:
            self.head = head
            self.isSingle = False
            self.target = target
            self.model = model
            self.__create_dep_graph()
            self.namedArgumentList = {}
            self.argumentList = []
            self.predicate = head + "_" + target
            self.path = self.__shortest_path()
        
            pos = 0        
            for node in self.path:
                attrs = get_attr_list_from_model(model, node)
                for a in attrs:
                    self.argumentList.append(self.predicate + "." + a)
                    self.namedArgumentList[self.predicate + "." + a] = pos
                    pos += 1
            self.argCount = pos
        else:
            self.predicate = predicate
            self.isSingle = True
            self.argumentList = copy.copy(get_attr_list_from_model(model, predicate))
            self.argCount = len(self.argumentList)
            self.namedArgumentList = {}
            self.target = None
            self.head = None
            self.model = model

            i = 0
            for node in self.argumentList:
                self.namedArgumentList["%s.%s" % (predicate, node)] = i
                i += 1
            
    
    def __create_dep_graph(self):
        """Create a graph with the dependencies of the model"""
        self.g = Graph()
        for cd in self.model['classDefs']:
            # Add a node for every class
            self.g.add_node(cd['name'])
        
        for e in self.model['dependencyDefs']:
            self.g.add_edge(e[0], e[1])
        
        # print >> sys.stderr, "Number nodes: %d" % number_of_nodes(g)
        # print >> sys.stderr, "Number edges: %d" % number_of_edges(g)
        return self.g
    
    def __shortest_path(self):
        """Search the shortest path between head and target"""
        return shortest_path(self.g, self.head, self.target)
    
    def get_arg_count(self):
        """get the number of arguments of the predicate"""
        return self.argCount
        
    def get_position(self, attrName):
        """Get the position of the named attribute"""
        i = self.namedArgumentList[attrName]
        if i == None:
            i = self.namedArgumentList[self.predicate + "." + attrName]
            if i == None:
                return 0
        return i
    
    def get_attribute_name(self, pos):
        """Get the name of the attribute at pos"""
        return self.argumentList[pos]
