import sys, marshal

from Ft.Xml.Domlette import NonvalidatingReader, Element
from Ft.Lib import Uri

# Parse the given rules and store them in the repository
class RuleParser(object):
    # @AttributeType string
    # The default storage.
    ___STORAGE = 'data/db.rules'

    
    # @ReturnType void
    # @ParamType aFilename 
    # Parse a given set of rules in the xml file.
    @staticmethod
    def parse(aFilename, config=None):
        """Parse the rules defined in the xml to the python
        data structure and marshal it to ___STORAGE
        
        The data structure is a list of rules. Rules are hashes
        that have the keys name, type, priority and message. As
        well as keys for subrules and params which are lists with 
        the subsequent entries.
        """
        
        if config == None:
            config = ConfigParser.ConfigParser()
            config.readfp(open('rubacon/defaults.cfg'))
            logging.debug("Created config parser in informationParser.")
            
        RuleParser.___STORAGE = config.get('rules', 'storage')
        
        
        file_uri = Uri.OsPathToUri(aFilename)
        doc = NonvalidatingReader.parseUri(file_uri)
        
        
        # Iterate over the rules in the file
        rules = doc.xpath('//rule')
        rs = []
        for rule in rules:
            r = {}
            subrules = []
            r['name'] = rule.attributes[(None, 'name')].value
            r['type'] = rule.attributes[(None, 'type')].value
            r['priority'] = rule.attributes[(None, 'priority')].value
            for child in rule.childNodes:
                #print >> sys.stderr, child
                # Skip the whitespace text nodes
                if not isinstance(child, Element):
                    continue
                if child.localName == 'message':
                    r['message'] = child.firstChild.nodeValue
                elif child.localName == 'subrule':
                    # print >> sys.stderr, child
                    subrule = {}
                    constraints = []
                    subrule['name'] = child.attributes[(None, 'name')].value
                    for sc in child.childNodes:
                        constraint = {}
                        if not isinstance(sc, Element):
                            continue
                        if sc.localName == 'head':
                            subrule['head'] = sc.attributes[(None, 'element')].value
                        elif sc.localName == 'target':
                            subrule['target'] = sc.attributes[(None, 'element')].value
                        elif sc.localName == 'constraint':
                            entries = []
                            constraint['name'] = sc.attributes[(None, 'name')].value
                            constraint['type'] = sc.attributes[(None, 'type')].value
                            for cons in sc.childNodes:
                                entry = {}
                                if not isinstance(cons, Element):
                                    continue
                                if cons.localName == 'reference':
                                    entry['type'] = 'reference'
                                    entry['name'] = cons.attributes[(None, 'name')].value
                                if cons.localName == 'parameter':
                                    entry['type'] = 'parameter'
                                    entry['name'] = cons.attributes[(None, 'name')].value
                                entries.append(entry)
                            constraint['entries'] = entries
                            constraints.append(constraint)
                        subrule['constraints'] = constraints
                    subrules.append(subrule)
                elif child.localName == 'param':
                    if not 'params' in r:
                        r['params'] = []
                    
                    r['params'].append(child.attributes[(None, 'name')].value)
                        
                else:
                    print >> sys.stderr, "Unknown element: %s" % child.localName
                # print >> sys.stderr, child.localName
                r['subrules'] = subrules
            #print >> sys.stderr, r
            rs.append(r)
        # That's all, marshal to a file now:
        ouf = open(RuleParser.___STORAGE, 'wb')
        marshal.dump(rs, ouf)
        ouf.close()
        

    # Remove the rules from the repository.
    @staticmethod
    def clearStorage():
        remove(RuleParser.___STORAGE)
    
    @staticmethod
    def getRulesFromStore():
        """Read the rules from the store"""
        inf = open(RuleParser.___STORAGE)
        r = marshal.load(inf)
        inf.close()
        return r
