# This parser reads the definitions from the XMI data.
#
# TODO: we need to check with which case tools it will work. Currently we use ArgoUML and its XMI export for the tool.

# from ..backend.XMLStorage import XMLStorage

from Ft.Xml import Parse
from Ft.Xml.XPath.Context import Context
from Ft.Xml.XPath import Evaluate

import marshal
import sys, logging, ConfigParser

# TODO: make the method static class methods
class ModelParser(object):
    # @AttributeType string
    # The default storage for models.
    ___STORAGE = 'data/db.model'
    
    # @ReturnType void
    # @ParamType aFilename string
    # Parse the XMI file to the given collection. The collection normally is db.model
    @staticmethod
    def parse(aFilename, config=None):
        """Parse the given XMI to the storage"""
        if config == None:
            config = ConfigParser.ConfigParser()
            config.readfp(open('rubacon/defaults.cfg'))
            logging.debug("Created config parser in modelParser.")
        ModelParser.___STORAGE = config.get('model', 'storage')
        
        logging.debug("Create UMLClassReader.")   
        ucr = UMLClassReader()
        classDefs = ucr.parse(aFilename)
        
        # -> Converting this to XML is mere overkill
        # -> and only good for exploit processor capacity!
        # -> We marshal this to a file
        ouf = open(ModelParser.___STORAGE, 'wb')
        marshal.dump(classDefs, ouf)
        ouf.close()

class UMLClassReader(object):
    """Extract class definitions from a XMI file"""
    
    # TODO: Extract this is a config parameter, so that it can
    # easily be adapted for a variety of case tools.
    # Known to work so far: ArgoUML, VP-UML
    ___NAMESPACE = u'org.omg.xmi.namespace.UML'
    
    def __init__(self):
        self.uml_classes = []
    
    def parse(self, name):
        """Extract the list of classes"""
        doc = Parse(name)
        # Namespace declaration:
        NSS = {u'Uml': self.___NAMESPACE}
        ctx = Context(doc, processorNss=NSS)
        defList = []
        classNameDict = {}    # Save the id -> name mapping for the dep-graph
        umlDefs = Evaluate(u'//Uml:Class', ctx)
        #print >> sys.stderr, "UMLDefs: " + doc.toprettyxml()
        # print >> sys.stderr, umlDefs
        for e in umlDefs:
            # print >> sys.stderr, e.localName
            if (None, u'name') in e.attributes:
                n = e.attributes[(None, u'name')].value
                logging.debug("Parsing class '%s'." % n)
                i = e.attributes[(None, u'xmi.id')].value
                classNameDict[i] = n
                ctx2 = Context(None, processorNss=NSS)
                xp = u"//Uml:Class[@xmi.id='" + i + "']/*/Uml:Attribute"
                attrDefs = Evaluate(xp, e, ctx2)
                attrList = []
                for a in attrDefs:
                    attrList.append(a.attributes[(None, u'name')].value)
                defList.append({'name': n, 'xmi_id': i, 'attrs': attrList})
        deps = Evaluate(u'//Uml:Dependency', ctx)
        depList = []
        for dep in deps:
            #print >> sys.stderr, "Deps:"
            #print >> sys.stderr, dep
            #print >> sys.stderr, "ChildNodes:"
            #print >> sys.stderr, dep.childNodes
            if not dep.childNodes == []:
                ctx2 = Context(dep, processorNss=NSS)
                i = dep.attributes[(None, 'xmi.id')].value
                con_name = dep.attributes[(None, 'name')].value
                xp = u"//Uml:Dependency[@xmi.id='" + i + "']/Uml:Dependency.client"
                client_node = Evaluate(xp, ctx2) # [0].firstChild
                #print >> sys.stderr, "Client Node"
                #print >> sys.stderr, client_node[0].childNodes[1]
                client = client_node[0].childNodes[1].attributes[(None, u'xmi.idref')].value
                client = classNameDict[client]  # replace the id by the classname
                
                xp = u"//Uml:Dependency[@xmi.id='" + i + "']/Uml:Dependency.supplier"
                supplier_node = Evaluate(xp, ctx2)[0].childNodes[1]
                supplier = supplier_node.attributes[(None, u'xmi.idref')].value
                 #print >> sys.stderr, "idref: %s" % supplier
                supplier = classNameDict[supplier]   # Use the classname here
                
                #.getElementsByTagNameNS(self.___NAMESPACE, "Class")[0].getAttribute('xmi.idref')
                # supplier = dep.getAttribute('supplier')
                depList.append([client, supplier, con_name])
        # print >> sys.stderr, defList
        # print >> sys.stderr, depList
        return {'classDefs': defList, 'dependencyDefs': depList}
	
